use nom::error::{ErrorKind, ParseError};
use nom::{
	AsBytes, Compare, CompareResult, ExtendInto, FindSubstring, FindToken, IResult, InputIter,
	InputLength, InputTake, InputTakeAtPosition, Needed, Offset, ParseTo, Slice,
};
use std::ops::{Range, RangeFrom, RangeFull, RangeTo};
use std::str::{CharIndices, Chars, FromStr};

#[derive(Copy, Clone, Debug)]
pub struct TextSpan<'a> {
	value: &'a str,
	offset: usize,
	length: usize,
}

impl<'a> PartialEq for TextSpan<'a> {
	fn eq(&self, other: &Self) -> bool {
		self.as_ref().eq(other.as_ref())
	}
}

#[derive(Copy, Clone, Debug, Eq, PartialEq)]
pub struct TokenPosition {
	pub offset: usize,
	pub length: usize,
	pub line: usize,
	pub column: usize,
}

impl<'a> TextSpan<'a> {
	pub fn new(inp: &'a str) -> Self {
		Self {
			value: inp,
			offset: 0,
			length: inp.len(),
		}
	}

	pub fn offset(&self) -> usize {
		self.offset
	}

	pub fn len(&self) -> usize {
		self.length
	}

	pub fn calculate_position(&self) -> TokenPosition {
		let mut current_line = 0;
		let mut total_chars = 0;

		for line in self.value.lines() {
			current_line += 1;

			if (total_chars + line.len()) >= self.offset {
				return TokenPosition {
					line: current_line,
					offset: self.offset,
					length: self.length,
					column: self.offset - total_chars,
				};
			}

			total_chars += line.len();
		}

		TokenPosition {
			line: current_line,
			offset: self.offset,
			length: self.length,
			column: self.value.lines().last().map(|li| li.len()).unwrap_or(0),
		}
	}

	pub fn advance_self_by(&mut self, amount: usize) {
		self.offset += amount;
	}

	pub fn advance_by(&self, amount: usize) -> Self {
		Self {
			offset: self.offset + amount,
			..*self
		}
	}

	pub fn into_value(self) -> &'a str {
		&self.value[self.offset..(self.offset + self.length)]
	}
}

impl<'a> AsRef<str> for TextSpan<'a> {
	fn as_ref(&self) -> &'a str {
		self.value
			.slice(self.offset..self.value.len().min(self.offset + self.length))
	}
}

impl<'a> AsBytes for TextSpan<'a> {
	fn as_bytes(&self) -> &[u8] {
		self.as_ref().as_bytes()
	}
}

impl<'a> Compare<&'a str> for TextSpan<'a> {
	fn compare(&self, t: &'a str) -> CompareResult {
		self.as_ref().compare(t)
	}

	fn compare_no_case(&self, t: &'a str) -> CompareResult {
		self.as_ref().compare_no_case(t)
	}
}

impl<'a> ExtendInto for TextSpan<'a> {
	type Item = char;
	type Extender = String;

	fn new_builder(&self) -> Self::Extender {
		self.as_ref().new_builder()
	}

	fn extend_into(&self, acc: &mut Self::Extender) {
		self.as_ref().extend_into(acc)
	}
}

impl<'a> FindSubstring<&'a str> for TextSpan<'a> {
	fn find_substring(&self, substr: &'a str) -> Option<usize> {
		self.as_ref().find_substring(substr)
	}
}

impl<'a> FindToken<&'a u8> for TextSpan<'a> {
	fn find_token(&self, token: &'a u8) -> bool {
		self.as_ref().find_token(token)
	}
}

impl<'a> InputIter for TextSpan<'a> {
	type Item = char;
	type Iter = CharIndices<'a>;
	type IterElem = Chars<'a>;

	fn iter_indices(&self) -> Self::Iter {
		self.into_value().iter_indices()
	}

	fn iter_elements(&self) -> Self::IterElem {
		self.into_value().iter_elements()
	}

	fn position<P>(&self, predicate: P) -> Option<usize>
	where
		P: Fn(Self::Item) -> bool,
	{
		self.as_ref().position(predicate)
	}

	fn slice_index(&self, count: usize) -> Result<usize, Needed> {
		self.as_ref().slice_index(count)
	}
}

impl<'a> InputLength for TextSpan<'a> {
	fn input_len(&self) -> usize {
		self.as_ref().input_len()
	}
}

impl<'a> InputTake for TextSpan<'a> {
	fn take(&self, count: usize) -> Self {
		Self {
			offset: self.offset,
			value: self.value,
			length: count,
		}
	}

	fn take_split(&self, count: usize) -> (Self, Self) {
		(
			Self {
				value: self.value,
				offset: self.offset,
				length: count,
			},
			Self {
				value: self.value,
				offset: self.offset.saturating_add(count),
				length: self.length.saturating_sub(count),
			},
		)
	}
}

impl<'a> InputTakeAtPosition for TextSpan<'a> {
	type Item = char;

	fn split_at_position<P, E: ParseError<Self>>(&self, predicate: P) -> IResult<Self, Self, E>
	where
		P: Fn(Self::Item) -> bool,
	{
		match self.as_ref().find(predicate) {
			Some(idx) => {
				let (found, remainder) = self.take_split(idx);
				Ok((remainder, found))
			}
			None => Err(nom::Err::Incomplete(Needed::new(1))),
		}
	}

	fn split_at_position1<P, E: ParseError<Self>>(
		&self,
		predicate: P,
		e: ErrorKind,
	) -> IResult<Self, Self, E>
	where
		P: Fn(Self::Item) -> bool,
	{
		match self.as_ref().find(predicate) {
			Some(0) => Err(nom::Err::Error(E::from_error_kind(*self, e))),
			Some(idx) => {
				let (found, remainder) = self.take_split(idx);
				Ok((remainder, found))
			}
			None => Err(nom::Err::Incomplete(Needed::new(1))),
		}
	}

	fn split_at_position_complete<P, E: ParseError<Self>>(
		&self,
		predicate: P,
	) -> IResult<Self, Self, E>
	where
		P: Fn(Self::Item) -> bool,
	{
		match self.as_ref().find(predicate) {
			Some(idx) => {
				let (found, remainder) = self.take_split(idx);
				Ok((remainder, found))
			}
			None => Ok((
				Self {
					value: self.value,
					offset: self.value.len(),
					length: 0,
				},
				*self,
			)),
		}
	}

	fn split_at_position1_complete<P, E: ParseError<Self>>(
		&self,
		predicate: P,
		e: ErrorKind,
	) -> IResult<Self, Self, E>
	where
		P: Fn(Self::Item) -> bool,
	{
		match self.as_ref().find(predicate) {
			Some(0) => Err(nom::Err::Error(E::from_error_kind(*self, e))),
			Some(idx) => {
				let (found, remainder) = self.take_split(idx);
				Ok((remainder, found))
			}
			None => {
				if self.length == 0 {
					Err(nom::Err::Error(E::from_error_kind(*self, e)))
				} else {
					Ok((
						Self {
							value: self.value,
							offset: self.value.len(),
							length: 0,
						},
						*self,
					))
				}
			}
		}
	}
}

impl<'a> Offset for TextSpan<'a> {
	fn offset(&self, second: &Self) -> usize {
		self.offset.saturating_sub(second.offset)
	}
}

impl<'a, R: FromStr> ParseTo<R> for TextSpan<'a> {
	fn parse_to(&self) -> Option<R> {
		self.as_ref().parse().ok()
	}
}

impl<'a> nom::Slice<Range<usize>> for TextSpan<'a> {
	fn slice(&self, range: Range<usize>) -> Self {
		let start = range.start;
		let length = range.end - range.start;
		Self {
			value: self.value,
			offset: self.offset + start,
			length,
		}
	}
}
impl<'a> nom::Slice<RangeTo<usize>> for TextSpan<'a> {
	fn slice(&self, range: RangeTo<usize>) -> Self {
		Self {
			value: self.value,
			offset: self.offset,
			length: range.end,
		}
	}
}
impl<'a> nom::Slice<RangeFrom<usize>> for TextSpan<'a> {
	fn slice(&self, range: RangeFrom<usize>) -> Self {
		let start = range.start;
		let offset = self.offset + start;
		let length = self.value.len() - offset;

		Self {
			value: self.value,
			offset,
			length,
		}
	}
}
impl<'a> nom::Slice<RangeFull> for TextSpan<'a> {
	fn slice(&self, _range: RangeFull) -> Self {
		*self
	}
}
