mod _span;
mod atoms;
mod keywords;
mod operators;
mod primitives;
mod strings;
mod tokens;

use keywords::{
	token_alias, token_else, token_export, token_finally, token_for, token_from, token_function,
	token_if, token_import, token_let, token_null, token_print, token_return, token_struct,
	token_super, token_this, token_typeof, token_while,
};
use operators::{
	token_asterisk, token_bang, token_bang_equal, token_caret, token_colon, token_comma, token_dot,
	token_double_ampersand, token_double_pipe, token_equal, token_equal_equal, token_greater,
	token_greater_equal, token_left_brace, token_left_paren, token_less, token_less_equal,
	token_minus, token_modulo, token_plus, token_right_brace, token_right_paren, token_semicolon,
	token_slash,
};
use primitives::{token_boolean, token_float, token_ident, token_int};
use strings::token_string;
pub use tokens::{ScriptToken, ScriptTokenType};

use crate::error::TokenError;
pub(crate) use atoms::{raw_decimal, raw_false, raw_true, Span};

mod _lex {
	use super::*;
	use nom::branch::alt;
	use nom::IResult;

	use nom::character::complete::multispace0;
	use nom::multi::fold_many0;
	use nom::sequence::delimited;

	fn any_token(input: Span) -> IResult<Span, ScriptToken> {
		alt((
			alt((
				token_if,
				token_function,
				token_alias,
				token_for,
				token_let,
				token_else,
				token_this,
				token_null,
				token_while,
				token_return,
				token_print,
				token_export,
				token_import,
				token_struct,
				token_super,
				token_from,
				token_typeof,
				token_finally,
			)),
			// Double character tokens
			alt((
				token_equal_equal,
				token_bang_equal,
				token_less_equal,
				token_greater_equal,
				token_double_ampersand,
				token_double_pipe,
			)),
			alt((
				token_plus,
				token_minus,
				token_asterisk,
				token_slash,
				token_bang,
				token_comma,
				token_dot,
				token_caret,
				token_modulo,
				token_left_brace,
				token_left_paren,
				token_right_brace,
				token_right_paren,
				token_semicolon,
				token_colon,
			)),
			alt((token_less, token_greater, token_equal)),
			alt((
				token_float,
				token_int,
				token_boolean,
				token_string,
				token_ident,
			)),
		))(input)
	}

	pub fn script_to_tokens<'a, 'b: 'a>(
		script: &'b str,
	) -> Result<Vec<ScriptToken<'a>>, TokenError<'a>> {
		let script_span = Span::new(script);
		let (_, tokens) = fold_many0(
			delimited(multispace0, any_token, multispace0),
			Vec::new,
			|mut list, tok| {
				list.push(tok);
				list
			},
		)(script_span)?;

		Ok(tokens)
	}

	#[cfg(test)]
	mod lex_test {
		use super::*;
		use test_case::test_case;

		#[test_case("%" => Ok(ScriptTokenType::Modulo) ; "Expects ScriptTokenType::Modulo")]
		#[test_case("^" => Ok(ScriptTokenType::Caret) ; "Expects ScriptTokenType::Caret")]
		#[test_case("!" => Ok(ScriptTokenType::Bang) ; "Expects ScriptTokenType::Bang")]
		#[test_case("(" => Ok(ScriptTokenType::LeftParen) ; "Expects ScriptTokenType::LeftParen")]
		#[test_case(")" => Ok(ScriptTokenType::RightParen) ; "Expects ScriptTokenType::RightParen")]
		#[test_case("{" => Ok(ScriptTokenType::LeftBrace) ; "Expects ScriptTokenType::LeftBrace")]
		#[test_case("}" => Ok(ScriptTokenType::RightBrace) ; "Expects ScriptTokenType::RightBrace")]
		#[test_case("," => Ok(ScriptTokenType::Comma) ; "Expects ScriptTokenType::Comma")]
		#[test_case("." => Ok(ScriptTokenType::Dot) ; "Expects ScriptTokenType::Dot")]
		#[test_case("-" => Ok(ScriptTokenType::Minus) ; "Expects ScriptTokenType::Minus")]
		#[test_case("+" => Ok(ScriptTokenType::Plus) ; "Expects ScriptTokenType::Plus")]
		#[test_case(";" => Ok(ScriptTokenType::Semicolon) ; "Expects ScriptTokenType::Semicolon")]
		#[test_case("/" => Ok(ScriptTokenType::Slash) ; "Expects ScriptTokenType::Slash")]
		#[test_case("*" => Ok(ScriptTokenType::Asterisk) ; "Expects ScriptTokenType::Asterisk")]
		#[test_case("=" => Ok(ScriptTokenType::Equal) ; "Expects ScriptTokenType::Equal")]
		#[test_case("!=" => Ok(ScriptTokenType::BangEqual) ; "Expects ScriptTokenType::BangEqual")]
		#[test_case("==" => Ok(ScriptTokenType::EqualEqual) ; "Expects ScriptTokenType::EqualEqual")]
		#[test_case(">" => Ok(ScriptTokenType::Greater) ; "Expects ScriptTokenType::Greater")]
		#[test_case(">=" => Ok(ScriptTokenType::GreaterEqual) ; "Expects ScriptTokenType::GreaterEqual")]
		#[test_case("&&" => Ok(ScriptTokenType::DoubleAmpersand) ; "Expects ScriptTokenType::DoubleAmpersand")]
		#[test_case("||" => Ok(ScriptTokenType::DoublePipe) ; "Expects ScriptTokenType::DoublePipe")]
		#[test_case("<" => Ok(ScriptTokenType::Less) ; "Expects ScriptTokenType::Less")]
		#[test_case("<=" => Ok(ScriptTokenType::LessEqual) ; "Expects ScriptTokenType::LessEqual")]
		#[test_case("\"Foo\"" => Ok(ScriptTokenType::OwnedString(String::from("Foo"))) ; "Expects ScriptTokenType::OwnedString")]
		#[test_case("123" => Ok(ScriptTokenType::Integer(123)) ; "Expects ScriptTokenType::Integer")]
		#[test_case("1_2_3" => Ok(ScriptTokenType::Integer(123)) ; "Expects ScriptTokenType::Integer underscored")]
		#[test_case("123.123" => Ok(ScriptTokenType::Float(123.123)) ; "Expects ScriptTokenType::Float")]
		#[test_case("12_3.1_23" => Ok(ScriptTokenType::Float(123.123)) ; "Expects ScriptTokenType::Float underscored")]
		#[test_case("struct" => Ok(ScriptTokenType::Class) ; "Expects ScriptTokenType::Class")]
		#[test_case("else" => Ok(ScriptTokenType::Else) ; "Expects ScriptTokenType::Else")]
		#[test_case("fn" => Ok(ScriptTokenType::Function) ; "Expects ScriptTokenType::Function")]
		#[test_case("for" => Ok(ScriptTokenType::For) ; "Expects ScriptTokenType::For")]
		#[test_case("if" => Ok(ScriptTokenType::If) ; "Expects ScriptTokenType::If")]
		#[test_case("null" => Ok(ScriptTokenType::Null) ; "Expects ScriptTokenType::Null")]
		#[test_case("print" => Ok(ScriptTokenType::Print) ; "Expects ScriptTokenType::Print")]
		#[test_case("return" => Ok(ScriptTokenType::Return) ; "Expects ScriptTokenType::Return")]
		#[test_case("super" => Ok(ScriptTokenType::Super) ; "Expects ScriptTokenType::Super")]
		#[test_case("this" => Ok(ScriptTokenType::This) ; "Expects ScriptTokenType::This")]
		#[test_case("let" => Ok(ScriptTokenType::Let) ; "Expects ScriptTokenType::Let")]
		#[test_case("while" => Ok(ScriptTokenType::While) ; "Expects ScriptTokenType::While")]
		#[test_case("export" => Ok(ScriptTokenType::Export) ; "Expects ScriptTokenType::Export")]
		#[test_case("import" => Ok(ScriptTokenType::Import) ; "Expects ScriptTokenType::Import")]
		#[test_case("as" => Ok(ScriptTokenType::Alias) ; "Expects ScriptTokenType::Alias")]
		#[test_case("from" => Ok(ScriptTokenType::From) ; "Expects ScriptTokenType::From")]
		#[test_case("typeof" => Ok(ScriptTokenType::Typeof) ; "Expects ScriptTokenType::Typeof")]
		#[test_case("finally" => Ok(ScriptTokenType::Finally) ; "Expects ScriptTokenType::Finally")]
		#[test_case("foo" => Ok(ScriptTokenType::Identifier(String::from("foo"))) ; "Expects ScriptTokenType::Identifier foo")]
		#[test_case("foo_bar" => Ok(ScriptTokenType::Identifier(String::from("foo_bar"))) ; "Expects ScriptTokenType::Identifier foo_bar")]
		#[test_case("fo2ob4r" => Ok(ScriptTokenType::Identifier(String::from("fo2ob4r"))) ; "Expects ScriptTokenType::Identifier fo2ob4r")]
		#[test_case("foobar" => matches Ok(ScriptTokenType::Identifier(_)))]
		fn correct_lexing(inp: &str) -> Result<ScriptTokenType, ()> {
			any_token(Span::new(inp))
				.map_err(|_| ())
				.map(|tok| tok.1.token_type)
		}
	}
}

pub use _lex::script_to_tokens;
