use crate::lexer::Span;
use crate::lexer::{ScriptToken, ScriptTokenType};
use crate::parse_ops;
use nom::bytes::complete::tag;
use nom::IResult;
use nom_locate::position;

parse_ops!(
	tag ;
	(token_colon, ":" => ScriptTokenType::Colon),
	(token_left_paren, "(" => ScriptTokenType::LeftParen),
	(token_right_paren, ")" => ScriptTokenType::RightParen),
	(token_left_brace, "{" => ScriptTokenType::LeftBrace),
	(token_right_brace, "}" => ScriptTokenType::RightBrace),
	(token_comma, "," => ScriptTokenType::Comma),
	(token_dot, "." => ScriptTokenType::Dot),
	(token_minus, "-" => ScriptTokenType::Minus),
	(token_plus, "+" => ScriptTokenType::Plus),
	(token_semicolon, ";" => ScriptTokenType::Semicolon),
	(token_slash, "/" => ScriptTokenType::Slash),
	(token_asterisk, "*" => ScriptTokenType::Asterisk),
	(token_bang, "!" => ScriptTokenType::Bang),
	(token_bang_equal, "!=" => ScriptTokenType::BangEqual),
	(token_equal, "=" => ScriptTokenType::Equal),
	(token_equal_equal, "==" => ScriptTokenType::EqualEqual),
	(token_greater, ">" => ScriptTokenType::Greater),
	(token_greater_equal, ">=" => ScriptTokenType::GreaterEqual),
	(token_less, "<" => ScriptTokenType::Less),
	(token_less_equal, "<=" => ScriptTokenType::LessEqual),
	(token_double_pipe, "||" => ScriptTokenType::DoublePipe),
	(token_double_ampersand, "&&" => ScriptTokenType::DoubleAmpersand),
	(token_modulo, "%" => ScriptTokenType::Modulo),
	(token_caret, "^" => ScriptTokenType::Caret)
);

#[cfg(test)]
mod operator_checks {
	use super::*;
	use crate::lexer::Span;
	use test_case::test_case;

	fn s(st: &str) -> Span {
		Span::new(st)
	}

	#[test_case(token_colon, ":" => Ok(ScriptTokenType::Colon) ; "Scan token_colon")]
	#[test_case(token_left_paren, "(" => Ok(ScriptTokenType::LeftParen) ; "Scan token_left_paren")]
	#[test_case(token_right_paren, ")" => Ok(ScriptTokenType::RightParen) ; "Scan token_right_paren")]
	#[test_case(token_left_brace, "{" => Ok(ScriptTokenType::LeftBrace) ; "Scan token_left_brace")]
	#[test_case(token_right_brace, "}" => Ok(ScriptTokenType::RightBrace) ; "Scan token_right_brace")]
	#[test_case(token_comma, "," => Ok(ScriptTokenType::Comma) ; "Scan token_comma")]
	#[test_case(token_dot, "." => Ok(ScriptTokenType::Dot) ; "Scan token_dot")]
	#[test_case(token_minus, "-" => Ok(ScriptTokenType::Minus) ; "Scan token_minus")]
	#[test_case(token_plus, "+" => Ok(ScriptTokenType::Plus) ; "Scan token_plus")]
	#[test_case(token_semicolon, ";" => Ok(ScriptTokenType::Semicolon) ; "Scan token_semicolon")]
	#[test_case(token_slash, "/" => Ok(ScriptTokenType::Slash) ; "Scan token_slash")]
	#[test_case(token_asterisk, "*" => Ok(ScriptTokenType::Asterisk) ; "Scan token_asterisk")]
	#[test_case(token_bang, "!" => Ok(ScriptTokenType::Bang) ; "Scan token_bang")]
	#[test_case(token_bang_equal, "!=" => Ok(ScriptTokenType::BangEqual) ; "Scan token_bang_equal")]
	#[test_case(token_equal, "=" => Ok(ScriptTokenType::Equal) ; "Scan token_equal")]
	#[test_case(token_equal_equal, "==" => Ok(ScriptTokenType::EqualEqual) ; "Scan token_equal_equal")]
	#[test_case(token_greater, ">" => Ok(ScriptTokenType::Greater) ; "Scan token_greater")]
	#[test_case(token_greater_equal, ">=" => Ok(ScriptTokenType::GreaterEqual) ; "Scan token_greater_equal")]
	#[test_case(token_less, "<" => Ok(ScriptTokenType::Less) ; "Scan token_less")]
	#[test_case(token_less_equal, "<=" => Ok(ScriptTokenType::LessEqual) ; "Scan token_less_equal")]
	#[test_case(token_double_pipe, "||" => Ok(ScriptTokenType::DoublePipe) ; "Scan token_double_pipe")]
	#[test_case(token_double_ampersand, "&&" => Ok(ScriptTokenType::DoubleAmpersand) ; "Scan token_double_ampersand")]
	#[test_case(token_modulo, "%" => Ok(ScriptTokenType::Modulo) ; "Scan token_modulo")]
	#[test_case(token_caret, "^" => Ok(ScriptTokenType::Caret) ; "Scan token_caret")]
	fn parse_token(
		func: impl Fn(Span) -> IResult<Span, ScriptToken>,
		tok: &str,
	) -> Result<ScriptTokenType, nom::Err<nom::error::Error<Span>>> {
		func(s(tok)).map(|t| t.1.token_type)
	}
}
