use crate::lexer::{ScriptToken, ScriptTokenType, Span};
use nom::branch::alt;
use nom::bytes::complete::{is_not, take_while_m_n};
use nom::character::complete::{char as p_char, multispace1};
use nom::combinator::{map, map_opt, map_res, value, verify};
use nom::multi::fold_many1;
use nom::sequence::{delimited, preceded};
use nom::IResult;
use nom_locate::position;

/// Parse a unicode sequence, of the form u{XXXX}, where XXXX is 1 to 6
/// hexadecimal numerals. We will combine this later with parse_escaped_char
/// to parse sequences like \u{00AC}.
fn parse_unicode(input: Span) -> IResult<Span, char> {
	// `take_while_m_n` parses between `m` and `n` bytes (inclusive) that match
	// a predicate. `parse_hex` here parses between 1 and 6 hexadecimal numerals.
	let parse_hex = take_while_m_n(1, 6, |ch: char| ch.is_ascii_hexdigit());

	// `preceded` takes a prefix parser, and if it succeeds, returns the result
	// of the body parser. In this case, it parses u{XXXX}.
	let parse_delimited_hex = preceded(
		p_char::<Span, nom::error::Error<Span>>('u'),
		// `delimited` is like `preceded`, but it parses both a prefix and a suffix.
		// It returns the result of the middle parser. In this case, it parses
		// {XXXX}, where XXXX is 1 to 6 hex numerals, and returns XXXX
		delimited(p_char('{'), parse_hex, p_char('}')),
	);

	// `map_res` takes the result of a parser and applies a function that returns
	// a Result. In this case we take the hex bytes from parse_hex and attempt to
	// convert them to a u32.
	let parse_u32 = map_res(parse_delimited_hex, move |hex| {
		u32::from_str_radix(hex.fragment(), 16)
	});

	// map_opt is like map_res, but it takes an Option instead of a Result. If
	// the function returns None, map_opt returns an error. In this case, because
	// not all u32 values are valid unicode code points, we have to fallibly
	// convert to p_char with from_u32.
	let (span, char) = map_opt(parse_u32, char::from_u32)(input)?;
	Ok((span, char))
}

/// Parse an escaped character: \n, \t, \r, \u{00AC}, etc.
fn parse_escaped_char(input: Span) -> IResult<Span, char> {
	preceded(
		p_char('\\'),
		// `alt` tries each parser in sequence, returning the result of
		// the first successful match
		alt((
			parse_unicode,
			// The `value` parser returns a fixed value (the first argument) if its
			// parser (the second argument) succeeds. In these cases, it looks for
			// the marker characters (n, r, t, etc) and returns the matching
			// character (\n, \r, \t, etc).
			value('\n', p_char('n')),
			value('\r', p_char('r')),
			value('\t', p_char('t')),
			value('\u{08}', p_char('b')),
			value('\u{0C}', p_char('f')),
			value('\\', p_char('\\')),
			value('/', p_char('/')),
			value('"', p_char('"')),
		)),
	)(input)
}

/// Parse a backslash, followed by any amount of whitespace. This is used later
/// to discard any escaped whitespace.
fn parse_escaped_whitespace(input: Span) -> IResult<Span, Span> {
	preceded(p_char('\\'), multispace1)(input)
}

/// Parse a non-empty block of text that doesn't include \ or "
fn parse_literal(input: Span) -> IResult<Span, Span> {
	// `is_not` parses a string of 0 or more characters that aren't one of the
	// given characters.
	let not_quote_slash = is_not("\"\\");

	// `verify` runs a parser, then runs a verification function on the output of
	// the parser. The verification function accepts out output only if it
	// returns true. In this case, we want to ensure that the output of is_not
	// is non-empty.
	verify(not_quote_slash, |s: &Span| !s.fragment().is_empty())(input)
}

/// A string fragment contains a fragment of a string being parsed: either
/// a non-empty Literal (a series of non-escaped characters), a single
/// parsed escaped character, or a block of escaped whitespace.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum StringFragment<'a> {
	Literal(&'a str),
	EscapedChar(char),
	EscapedWS,
}

/// Combine parse_literal, parse_escaped_whitespace, and parse_escaped_char
/// into a StringFragment.
fn parse_fragment(input: Span) -> IResult<Span, StringFragment> {
	alt((
		map(parse_literal, |sp: Span| {
			StringFragment::Literal(sp.fragment())
		}),
		map(parse_escaped_char, StringFragment::EscapedChar),
		value(StringFragment::EscapedWS, parse_escaped_whitespace),
	))(input)
}

/// Parse a string. Use a loop of parse_fragment and push all of the fragments
/// into an output string.
pub fn token_string(input: Span) -> IResult<Span, ScriptToken> {
	// fold is the equivalent of iterator::fold. It runs a parser in a loop,
	// and for each output value, calls a folding function on each output value.
	let build_string = fold_many1(
		// Our parser function– parses a single string fragment
		parse_fragment,
		// Our init value, an empty string
		String::new,
		// Our folding function. For each fragment, append the fragment to the
		// string.
		|mut string, fragment| {
			match fragment {
				StringFragment::Literal(s) => string.push_str(s),
				StringFragment::EscapedChar(c) => string.push(c),
				StringFragment::EscapedWS => {}
			}
			string
		},
	);

	let (input, pos) = position(input)?;
	let (input, value) = delimited(p_char('"'), build_string, p_char('"'))(input)?;

	Ok((
		input,
		ScriptToken {
			position: pos,
			token_type: ScriptTokenType::OwnedString(value),
		},
	))
}

#[cfg(test)]
mod string_test {
	use super::token_string;
	use crate::lexer::Span;
	use crate::lexer::{ScriptToken, ScriptTokenType};
	use crate::map_result;

	#[test]
	fn parse_escaped_string() {
		let positive_cases = [
			(
				r#""This is an escaped String""#,
				String::from("This is an escaped String"),
			),
			(
				r#""This is an \"escaped\" String""#,
				String::from("This is an \"escaped\" String"),
			),
			(
				r#""Many whitespaces can be collapsed with a slash \
				and they won't matter""#,
				String::from(
					"Many whitespaces can be collapsed with a slash and they won't matter",
				),
			),
			(
				r#""Big whitespace preserved can be collapsed with a slash
				and they won't matter""#,
				String::from(
					r#"Big whitespace preserved can be collapsed with a slash
				and they won't matter"#,
				),
			),
		];

		for (program, expected) in positive_cases {
			map_result!(
				token_string(Span::new(program)),
				|_, token: ScriptToken| {
					assert_eq!(token.token_type, ScriptTokenType::OwnedString(expected))
				},
				program
			);
		}
	}
}
