use std::cmp::Ordering;
use std::fmt::{Display, Formatter};
use std::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Rem, RemAssign, Sub, SubAssign};

/// Represents a numerical value in a script
#[derive(Copy, Clone, Debug)]
#[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
#[cfg_attr(
	feature = "serde",
	serde(tag = "f", content = "v", rename_all = "snake_case")
)]
pub enum Number {
	Integer(i64),
	Float(f64),
}

impl Number {
	/// Create a new Number representing the provided integer
	#[inline]
	pub const fn integer(value: i64) -> Self {
		Self::Integer(value)
	}
	/// Create a new Number representing the provided float
	#[inline]
	pub const fn float(value: f64) -> Self {
		Self::Float(value)
	}
	/// Check if the internal representation of this number is a fractional value
	#[inline]
	pub const fn is_float(&self) -> bool {
		matches!(self, Number::Float(..))
	}
	/// Check if the internal representation of this number is an integer value
	#[inline]
	pub const fn is_integer(&self) -> bool {
		matches!(self, Number::Integer(..))
	}

	/// Create a copy of the value represented as an Integer internally.
	/// Will lose information if the value is a float and has a non-zero
	/// mantissa
	pub fn as_integer(&self) -> Self {
		match self {
			&Self::Float(value) => Self::Integer(value as i64),
			&Self::Integer(value) => Self::Integer(value),
		}
	}
	/// Create a copy of the value represented as a Float internally.
	/// Will not lose information in the conversion
	pub fn as_float(&self) -> Self {
		match self {
			&Self::Float(value) => Self::Float(value),
			&Self::Integer(value) => Self::Float(value as f64),
		}
	}

	/// Unwrap this number into a native rust value, represented as
	/// an integer
	/// Will lose information if the value is a float and has a non-zero
	/// mantissa
	pub fn as_i64(&self) -> i64 {
		match self {
			Self::Integer(val) => *val,
			Self::Float(val) => (*val) as i64,
		}
	}
	/// Unwrap this number into a native rust value, represented as
	/// a float
	/// Will not lose information in the conversion
	pub fn as_f64(&self) -> f64 {
		match self {
			Self::Integer(val) => (*val) as f64,
			Self::Float(val) => *val,
		}
	}

	/// Check to see if both value _and_ type matches between this and another
	/// Number.
	/// An escape hatch for situations where it is critical that values are (or
	/// aren't) the same internal type.
	pub fn matches(&self, other: Number) -> bool {
		use Number::*;
		match (*self, other) {
			(Integer(first), Integer(second)) => first == second,
			(Float(first), Float(second)) => first == second,
			_ => false,
		}
	}
}

impl PartialEq for Number {
	fn eq(&self, other: &Self) -> bool {
		use Number::*;
		match (self, other) {
			(Integer(first), Integer(second)) => first == second,
			(Float(first), Float(second)) => first == second,
			(Float(fl), Integer(int)) | (Integer(int), Float(fl)) => (*int as f64) == *fl,
		}
	}
}

impl PartialOrd for Number {
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		use Number::*;
		match (self, other) {
			(Integer(first), Integer(second)) => first.partial_cmp(second),
			(Float(first), Float(second)) => first.partial_cmp(second),
			(Float(first), Integer(second)) => first.partial_cmp(&(*second as f64)),
			(Integer(first), Float(second)) => (*first as f64).partial_cmp(second),
		}
	}
}

impl Add for Number {
	type Output = Number;
	fn add(self, rhs: Self) -> Self::Output {
		use Number::*;
		match (self, rhs) {
			(Integer(first), Integer(second)) => Integer(first.saturating_add(second)),
			(Float(first), Float(second)) => Float(first + second),
			(Float(fl), Integer(int)) | (Integer(int), Float(fl)) => Float(fl + int as f64),
		}
	}
}
impl AddAssign for Number {
	#[inline]
	fn add_assign(&mut self, rhs: Self) {
		*self = *self + rhs;
	}
}
impl Sub for Number {
	type Output = Number;
	fn sub(self, rhs: Self) -> Self::Output {
		use Number::*;
		match (self, rhs) {
			(Integer(first), Integer(second)) => Integer(first.saturating_sub(second)),
			(Float(first), Float(second)) => Float(first - second),
			(Float(first), Integer(second)) => Float(first - second as f64),
			(Integer(first), Float(second)) => Float(first as f64 - second),
		}
	}
}
impl SubAssign for Number {
	#[inline]
	fn sub_assign(&mut self, rhs: Self) {
		*self = *self - rhs;
	}
}
impl Mul for Number {
	type Output = Number;

	fn mul(self, rhs: Self) -> Self::Output {
		use Number::*;
		match (self, rhs) {
			(Integer(first), Integer(second)) => Integer(first * second),
			(Float(first), Float(second)) => Float(first * second),
			(Float(fl), Integer(int)) | (Integer(int), Float(fl)) => Float(fl * int as f64),
		}
	}
}
impl MulAssign for Number {
	#[inline]
	fn mul_assign(&mut self, rhs: Self) {
		*self = *self * rhs
	}
}
impl Div for Number {
	type Output = Number;
	#[inline]
	fn div(self, rhs: Self) -> Self::Output {
		Number::float(self.as_f64() / rhs.as_f64())
	}
}
impl DivAssign for Number {
	#[inline]
	fn div_assign(&mut self, rhs: Self) {
		*self = *self / rhs;
	}
}

impl Rem for Number {
	type Output = Number;
	fn rem(self, rhs: Self) -> Self::Output {
		use Number::*;
		match (self, rhs) {
			(Integer(first), Integer(second)) => Integer(first % second),
			(Float(first), Float(second)) => Float(first % second),
			(Float(first), Integer(second)) => Float(first % second as f64),
			(Integer(first), Float(second)) => Float(first as f64 % second),
		}
	}
}

impl RemAssign for Number {
	#[inline]
	fn rem_assign(&mut self, rhs: Self) {
		*self = *self % rhs;
	}
}

impl Display for Number {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		match self {
			Number::Float(val) => write!(f, "{}", val),
			Number::Integer(val) => write!(f, "{}", val),
		}
	}
}

macro_rules! impl_into_primitive {
    ($($type:ty),+) => {
		$(
		impl From<Number> for $type {
			fn from(value: Number) -> Self {
				match value {
					Number::Integer(value) => value as $type,
					Number::Float(value) => value as $type,
				}
			}
		}
		)+
	};
}

macro_rules! impl_from {
    (int: $($type:ty),+) => {
		$(
			impl From<$type> for Number {
				fn from(value: $type) -> Self {
					Number::Integer(value as i64)
				}
			}
		)+
	};
	(float: $($type:ty),+) => {
		$(
			impl From<$type> for Number {
				fn from(value: $type) -> Self {
					Number::Float(value as f64)
				}
			}
		)+
	};
}

impl_into_primitive!(u8, u16, u32, u64, usize, i8, i16, i32, i64, isize, f32, f64);
impl_from!(int: u8, u16, u32, u64, usize, i8, i16, i32, i64, isize);
impl_from!(float: f32, f64);

#[cfg(test)]
mod gcse_maths {
	use super::Number;
	use test_case::test_case;

	#[test]
	fn eq_ignores_internal() {
		assert_eq!(Number::integer(123), Number::float(123.0));
		assert_eq!(Number::integer(123), Number::integer(123));
		assert_eq!(Number::float(123.0), Number::float(123.0));
	}

	#[test_case(Number::integer(123), Number::integer(456) => Number::integer(579))]
	#[test_case(Number::float(123.0), Number::integer(456) => Number::float(579.0))]
	#[test_case(Number::float(123.0), Number::float(456.0) => Number::float(579.0))]
	#[test_case(Number::integer(123), Number::float(456.5) => Number::float(579.5))]
	fn addition(a: Number, b: Number) -> Number {
		a + b
	}

	#[test]
	fn type_coercion() {
		let first = Number::integer(150);
		let second = Number::float(5.0);

		assert!(first.is_integer());
		assert!(second.is_float());

		assert!((first + second).is_float(), "Addition did not create float");
		assert!(
			(first - second).is_float(),
			"Subtraction did not create float"
		);
		assert!((first * second).is_float(), "Multiply did not create float");
		assert!((first / second).is_float(), "Divide did not create float");

		assert!(
			(first / first).is_float(),
			"Divide should always create float"
		);
	}
}
