use crate::runtime::value::ForgeValue;
use crate::runtime::vm::const_data::{ConstData, ConstDataRef};
use crate::runtime::vm::opcode::OpCode;
use crate::utilities::Clown;
use std::fmt::{Display, Formatter};

#[derive(Default)]
pub struct Chunk {
	code: Vec<OpCode>,
	consts: ConstData,
}

impl From<Vec<OpCode>> for Chunk {
	fn from(value: Vec<OpCode>) -> Self {
		Self {
			code: value,
			consts: ConstData::default(),
		}
	}
}

pub struct ChunkRef<'scope> {
	code: &'scope [OpCode],
	consts: ConstDataRef<'scope>,
}

impl<'scope> From<&'scope Chunk> for ChunkRef<'scope> {
	fn from(value: &'scope Chunk) -> Self {
		Self {
			code: value.code.as_slice(),
			consts: ConstDataRef::from(value.consts.as_slice()),
		}
	}
}

impl Chunk {
	pub fn push_op(&mut self, op: OpCode) -> usize {
		let idx = self.code.len();
		self.code.push(op);
		idx
	}

	pub fn push_const_data(&mut self, data: ForgeValue) -> usize {
		let idx = self.consts.len();
		self.consts.push(data);
		idx
	}

	pub fn as_ref(&self) -> ChunkRef {
		ChunkRef {
			code: self.code.as_slice(),
			consts: self.consts.as_ref(),
		}
	}

	pub fn op_constant(&mut self, data: ForgeValue) {
		let idx = self.push_const_data(data);
		self.push_op(OpCode::Constant(idx));
	}
}

impl<'scope> Clown<Chunk> for ChunkRef<'scope> {
	fn clown(&self) -> Chunk {
		Chunk {
			consts: self.consts.clown(),
			code: Vec::from(self.code),
		}
	}
}

pub trait ChunkOps {
	fn as_slice(&self) -> &[OpCode];
	fn as_const_slice(&self) -> &[ForgeValue];

	fn as_chunk(&self) -> Chunk {
		ChunkRef {
			code: self.as_slice(),
			consts: self.as_const_slice().into(),
		}
		.clown()
	}

	fn write_chunk_data(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		let mut offset = 0;
		let slice = self.as_slice();

		while offset < slice.len() {
			let opcode = match slice.get(offset) {
				Some(value) => value,
				None => {
					break;
				}
			};

			write!(f, "[{:08x}] {} ", offset, opcode)?;
			opcode.format_operand(f)?;

			match opcode {
				OpCode::Constant(idx) => {
					write!(f, " {}", self.as_const_slice()[*idx])?;
				}
				_ => {}
			}

			writeln!(f)?;

			offset = opcode.next_offset(offset);
		}

		Ok(())
	}

	fn format_chunk(&self, name: impl Display) -> String;
}

macro_rules! chunk_ops_generic {
	() => {
		fn format_chunk(&self, name: impl Display) -> String {
			format!("== {} ==\n{}", name, self)
		}
	};
}

impl ChunkOps for Chunk {
	fn as_slice(&self) -> &[OpCode] {
		self.code.as_slice()
	}
	fn as_const_slice(&self) -> &[ForgeValue] {
		self.consts.as_slice()
	}

	chunk_ops_generic!();
}
impl<'scope> ChunkOps for ChunkRef<'scope> {
	fn as_slice(&self) -> &[OpCode] {
		self.code
	}
	fn as_const_slice(&self) -> &[ForgeValue] {
		*self.consts
	}

	chunk_ops_generic!();
}

impl Display for Chunk {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		self.write_chunk_data(f)
	}
}

impl<'scope> Display for ChunkRef<'scope> {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		self.write_chunk_data(f)
	}
}

#[cfg(test)]
mod _chunks {
	use super::*;

	#[test]
	fn basic_chunk_format() {
		let chunk = Chunk::from(vec![OpCode::Return]);
		let expected = "== test ==\n[00000000] op_return \n";

		assert_eq!(chunk.format_chunk("test"), String::from(expected));
	}

	#[test]
	fn const_chunk() {
		let mut chunk = Chunk::default();
		chunk.op_constant(ForgeValue::Null);
		chunk.push_op(OpCode::Return);

		let expected = format!(
			"== test ==\n[00000000] op_constant {:016x} null\n[00000001] op_return \n",
			0
		);

		assert_eq!(chunk.format_chunk("test"), expected);
	}
}
