use crate::lexer::ScriptTokenType;
use crate::pratt::{Scanner, ScannerError, ScannerErrorKind, ScannerIter, ScannerToken};

use crate::runtime::vm::Chunk;
use crate::runtime::vm::ParsingPrecedence;

#[derive(Clone, Default)]
pub struct ParsingState {
	previous: Option<ScannerToken>,
	current: Option<ScannerToken>,
}

impl ParsingState {
	pub fn with_current(token: ScannerToken) -> Self {
		Self {
			previous: None,
			current: Some(token),
		}
	}
	pub fn set_current(&mut self, token: Option<ScannerToken>) {
		self.previous = std::mem::take(&mut self.current);
		self.current = token;
	}
	pub fn was_previously(&self, tok: ScriptTokenType) -> bool {
		self.previous
			.as_ref()
			.map(|s| s.token == tok)
			.unwrap_or(false)
	}
	pub fn is_currently(&self, tok: ScriptTokenType) -> bool {
		self.current
			.as_ref()
			.map(|s| s.token == tok)
			.unwrap_or(false)
	}

	pub fn previous(&self) -> &Option<ScannerToken> {
		&self.previous
	}
	pub fn current(&self) -> &Option<ScannerToken> {
		&self.current
	}
}

pub struct Compiler<'a> {
	chunk: Chunk,
	source: ScannerIter<'a>,
	state: ParsingState,
}

type CompilerOp = Result<(), ScannerError>;

impl<'a> Compiler<'a> {
	pub fn new(source: &'a str) -> Compiler<'a> {
		Self {
			chunk: Chunk::default(),
			source: Scanner::new(source).into_iter(),
			state: ParsingState::default(),
		}
	}

	pub fn init_new(source: &'a str) -> Result<Compiler<'a>, ScannerError> {
		let mut cmp = Self::new(source);
		cmp.advance()?;
		Ok(cmp)
	}

	pub fn chunk(&self) -> &Chunk {
		&self.chunk
	}

	pub fn chunk_mut(&mut self) -> &mut Chunk {
		&mut self.chunk
	}

	pub fn compile(&mut self) -> Result<&Chunk, ScannerError> {
		self.compile_expression()?;
		self.assert_next(ScriptTokenType::Eof)?;
		Ok(self.chunk())
	}

	fn process_precedence(&mut self, _precedence: ParsingPrecedence) -> CompilerOp {
		Ok(())
	}

	fn compile_expression(&mut self) -> CompilerOp {
		compile_expression(self)
	}

	fn advance(&mut self) -> CompilerOp {
		if let Some(tok) = self.source.next() {
			self.state.set_current(Some(tok?));
		}
		Ok(())
	}

	fn assert_next(&mut self, next: ScriptTokenType) -> CompilerOp {
		self.advance()?;
		match self.current() {
			Some(val) => {
				if next == val.token {
					Ok(())
				} else {
					Err(ScannerError {
						kind: ScannerErrorKind::ExpectedDifferentToken {
							found: val.token.clone(),
							expected: vec![next],
						},
						position: self.source.position(),
					})
				}
			}
			None => Err(ScannerError {
				kind: ScannerErrorKind::UnexpectedEof,
				position: self.source.position(),
			}),
		}
	}

	fn discard_until(&mut self, until: ScriptTokenType) -> Result<(), ScannerError> {
		for tok in self.source.by_ref() {
			let tok = tok?;
			if until == tok.token {
				return Ok(());
			}
		}

		Err(ScannerError {
			kind: ScannerErrorKind::UnexpectedEof,
			position: self.source.position(),
		})
	}

	pub fn current(&self) -> &Option<ScannerToken> {
		&self.state.current
	}
	pub fn previous(&self) -> &Option<ScannerToken> {
		&self.state.previous
	}
}

fn compile_expression(_compiler: &mut Compiler) -> CompilerOp {
	Ok(())
}
