use crate::runtime::vm::Compiler;
use std::collections::HashMap;

#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Debug, Hash)]
pub enum ParsingRuleType {
	LeftParen,
	RightParen,
	LeftBrace,
	RightBrace,
	Comma,
	Dot,
	Semicolon,
	Bang,
	Minus,
	Asterisk,
	Slash,
	Plus,
	BangEqual,
	Equal,
	EqualEqual,
	Greater,
	GreaterEqual,
	Less,
	LessEqual,
	DoublePipe,
	DoubleAmpersand,
	Modulo,
	Caret,
	Identifier,
	String,
	OwnedString,
	Integer,
	Float,
	Boolean,
	Class,
	Else,
	Function,
	For,
	If,
	Null,
	Print,
	Return,
	Super,
	This,
	Let,
	While,
	Export,
	Import,
	Alias,
	From,
	Typeof,
	Finally,
}

#[repr(u8)]
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Debug, Hash)]
/// A given precedence level represents parsing for that level and all levels above it
pub enum ParsingPrecedence {
	None = 0,
	/// Direct assignment statements
	Assignment = 1,
	/// Boolean operators || and &&
	Boolean = 2,
	/// Equal / Not Equal operators == and !=
	Equality = 3,
	/// Ordering Comparison Operators >, <, <=, and >=
	Ordering = 4,
	/// Lower precedence binary operators + (add) and - (subtract)
	BinaryTerm = 5,
	/// Higher precedence binary operators * (multiply) and / (divide)
	BinaryFactor = 6,
	/// Unary operators ! (negation) and - (inversion)
	Unary = 7,
	/// Operators for data or process drilling . (property lookup) and `()` (function call)
	CallLookup = 8,
	Primary = 100,
}

pub struct ParsingRule {
	pub precedence: ParsingPrecedence,
	pub unary_rule: Option<Box<dyn Fn(&mut Compiler)>>,
	pub binary_rule: Option<Box<dyn Fn(&mut Compiler)>>,
}

pub type ParsingRules = HashMap<ParsingRuleType, ParsingRule>;

macro_rules! _inr_replace {
	($_t:tt $out:expr) => {
		$expr
	};
}
macro_rules! _inr_count {
    ($($tts:tt)*) => {<[()]>::len(&[$(_int_replace!($tts ())),*])};
}

#[macro_export]
macro_rules! rule_table {
    ($($token: expr, $unary_rule: expr, $binary_rule: expr, $precedence: expr),+) => {
		{
			let table = HashMap::with_capacity(_inr_count!($($token)*));
			$(
				table.insert($token, ParsingRule {
					precedence: $precedence,
					unary_rule: $unary_rule,
					binary_rule: $binary_rule,
				});
			)+;
			table
		}
	};
}
