#[macro_export]
macro_rules! deref_as {
    ($type:ty => $other:ty) => {
		impl std::ops::Deref for $type {
			type Target = $other;

			fn deref(&self) -> &Self::Target {
				&self.0
			}
		}
	};
    ($($lifetime:tt),+: $type:ty => $other:ty) => {
		impl <$($lifetime),+> std::ops::Deref for $type {
			type Target = $other;

			fn deref(&self) -> &Self::Target {
				&self.0
			}
		}
	};
    (mut $type:ty => $other:ty) => {
		deref_as!($type => $other);
		impl std::ops::DerefMut for $type {
			fn deref_mut(&mut self) -> &mut Self::Target {
				&mut self.0
			}
		}
    };
    ($($lifetime:tt),+ mut: $type:ty => $other:ty) => {
		deref_as!($($lifetime),+: $type => $other);
		impl <$($lifetime),+> std::ops::DerefMut for $type {
			fn deref_mut(&mut self) -> &mut Self::Target {
				&mut self.0
			}
		}
    };
}

#[macro_export]
macro_rules! parse_ops {
	(tag_ws ; $(($fn_name:ident, $tok:literal => $prod:expr)),+) => {
		$(pub fn $fn_name(input: Span) -> IResult<Span, ScriptToken> {
			let (input, pos) = position(input)?;
			let (input, _) = tag_ws($tok, input)?;
			Ok((
				input,
				ScriptToken {
					position: pos,
					token_type: $prod,
				},
			))
		})+
	};
	($fn: expr ; $(($fn_name:ident, $tok:literal => $prod:expr)),+) => {
		$(pub fn $fn_name(input: Span) -> IResult<Span, ScriptToken> {
			let (input, pos) = position(input)?;
			let (input, _) = $fn($tok)(input)?;
			Ok((
				input,
				ScriptToken {
					position: pos,
					token_type: $prod,
				},
			))
		})+
	};
}

/// "Clone-owned", clone a struct into an owned version of that struct, which will typically be
/// a different type to the implementor. If the struct is already an owned struct then the impl
/// _should_ perform a simple clone, though the impl may choose to perform otherwise
pub trait Clown<Target> {
	fn clown(&self) -> Target;
}

type LineNum = usize;
type ColumnNum = usize;

pub fn offset_to_line_column(source: &str, offset: usize) -> (LineNum, ColumnNum) {
	let mut remaining = offset;
	let mut current_line = 0;

	for line in source.lines() {
		current_line += 1;
		if remaining < line.len() {
			return (current_line, remaining + 1);
		}
		remaining -= line.len();
	}

	(
		current_line,
		source.lines().last().map(|l| l.len()).unwrap_or(0),
	)
}
